from typing import LiteralString
from urllib.parse import quote_plus

GOOGLE_MAPS_QUERY_STRING: LiteralString = "https://www.google.com/maps/search/{query}"


def gen_map_search_url(location: str) -> str:
    return GOOGLE_MAPS_QUERY_STRING.format(query=quote_plus(location))



from re import match as re_match
from typing import Literal

from pydantic import ConfigDict, Field, field_validator

from cpn_core.notifications.models.base import BaseNotificationEngineConfig


class DiscordNotificationEngineConfig(BaseNotificationEngineConfig):
    model_config = ConfigDict(
        title="Discord",
        frozen=True,
    )

    bot_token: str = Field(
        description="Bot token",
        examples=[
            "MTMzNzg4Ujq0NDI0NDYgNTcyMA.GpITQg.beoF9OxJScbKJwEz5Udy6bzrQJ8zI4BvndbaBA",
        ],
    )
    chat_id: int = Field(
        description="Chat ID",
        examples=[
            832930846182672436,
        ],
    )
    chat_type: Literal["user", "channel"]
    markdown: bool = Field(
        description="Gửi tin nhắn dạng markdown",
        default=True,
    )

    @field_validator("bot_token", mode="after")
    @classmethod
    def validate_bot_token(cls, _bot_token: str) -> str:
        if not re_match(
            r"^[A-Za-z0-9_\-]+\.[A-Za-z0-9_\-]+\.[A-Za-z0-9_\-]+$", _bot_token
        ):
            raise ValueError("Bot token is not valid")
        return _bot_token

    @field_validator("chat_id", mode="after")
    @classmethod
    def validate_user_id(cls, _user_id: int) -> int:
        if not re_match(r"^\d{18,19}$", f"{_user_id}"):
            raise ValueError("User id is not valid")
        return _user_id


__all__ = ["DiscordNotificationEngineConfig"]



from re import match as re_match

from pydantic import ConfigDict, Field, field_validator

from cpn_core.notifications.models.base import BaseNotificationEngineConfig


class TelegramNotificationEngineConfig(BaseNotificationEngineConfig):
    model_config = ConfigDict(
        title="Telegram",
        frozen=True,
    )

    bot_token: str = Field(
        description="Bot token Telegram",
        examples=[
            "2780473231:weiruAShGUUx4oLOMoUhd0GiREXSZcCq-uB",
        ],
    )
    chat_id: str = Field(
        description="Chat ID Telegram",
        examples=[
            "-1001790012349",
        ],
    )
    markdown: bool = Field(
        description="Gửi tin nhắn dạng markdown",
        default=True,
    )

    @field_validator("bot_token", mode="after")
    @classmethod
    def validate_bot_token(cls, _bot_token: str) -> str:
        if not re_match(r"^[0-9]+:.+$", _bot_token):
            raise ValueError("Bot token is not valid")
        return _bot_token

    @field_validator("chat_id", mode="after")
    @classmethod
    def validate_chat_id(cls, _chat_id: str) -> str:
        if not re_match(r"^[+-]?[0-9]+$", _chat_id):
            raise ValueError("Chat ID is not valid")
        return _chat_id


__all__ = ["TelegramNotificationEngineConfig"]



from logging import getLogger
from typing import override

from discord import (
    Client,
    DMChannel,
    Forbidden,
    GroupChannel,
    HTTPException,
    Intents,
    TextChannel,
    User,
)

from cpn_core.notifications.models.discord import DiscordNotificationEngineConfig

from .base import BaseNotificationEngine

logger = getLogger(__name__)


# FIXME: @NTGNguyen: fetch channel, id bla bla bla. The command_prefix seem bruh? not relate
class _DiscordNotificationCoreEngine:
    def __init__(
        self,
        discord: DiscordNotificationEngineConfig,
        messages: tuple[str, ...],
    ) -> None:
        self.discord: DiscordNotificationEngineConfig = discord
        self._messages: tuple[str, ...] = messages
        self._client = Client(intents=Intents.default())

    async def _send_channel(self) -> None:
        try:
            channel = await self._client.fetch_channel(self.discord.chat_id)
            if channel is None:
                logger.error(f"Discord channel ID {self.discord.chat_id}: Not found")
                return
            if (
                not isinstance(channel, TextChannel)
                or not isinstance(channel, GroupChannel)
                or not isinstance(channel, DMChannel)
            ):
                logger.error(
                    f"Discord channel ID {self.discord.chat_id}: Must be text channel"
                )
                return
            for message in self._messages:
                await channel.send(message)
            logger.info(f"Successfully sent to Discord channel: {self.discord.chat_id}")
        except Exception as e:
            logger.error(f"Discord channel ID {self.discord.chat_id}: {e}")

    async def _send_user(self) -> None:
        try:
            user: User = await self._client.fetch_user(self.discord.chat_id)
            for message in self._messages:
                await user.send(message)
            logger.info(f"Successfully sent to Discord user: {self.discord.chat_id}")
        except Forbidden as e:
            logger.error(
                f"Discord bot doesn't have permission to send to user {self.discord.chat_id}. {e}"
            )
        except HTTPException as e:
            logger.error(f"Failed to send message to {self.discord.chat_id}. {e}")
        except Exception as e:
            logger.error(
                f"Failed to send message to {self.discord.chat_id} (internal). {e}"
            )

    async def send(self) -> None:
        @self._client.event
        async def on_ready() -> None:  # pyright: ignore[reportUnusedFunction]
            match self.discord.chat_type:
                case "user":
                    await self._send_user()
                case "channel":
                    await self._send_channel()
            await self._client.close()

        await self._client.start(self.discord.bot_token)


class DiscordNotificationEngine(
    BaseNotificationEngine[DiscordNotificationEngineConfig]
):
    @override
    async def send(
        self,
        config: DiscordNotificationEngineConfig,
        messages: tuple[str, ...],
    ) -> None:
        discord_engine = _DiscordNotificationCoreEngine(config, messages)
        await discord_engine.send()



from enum import Enum


class ApiEnum(str, Enum):
    checkphatnguoi_vn = "checkphatnguoi.vn"
    csgt_vn = "csgt.vn"
    phatnguoi_vn = "phatnguoi.vn"
    zm_io_vn = "zm.io.vn"


__all__ = ["ApiEnum"]



from datetime import datetime
from http.cookies import SimpleCookie
from io import BytesIO
from logging import getLogger
from ssl import SSLContext
from ssl import create_default_context as ssl_create_context
from typing import Final, LiteralString, Self, override

from aiohttp import ClientError, ClientSession, ClientTimeout
from bs4 import BeautifulSoup, NavigableString, Tag
from PIL import Image
from pytesseract import image_to_string

from cpn_core.get_data.engines.base import BaseGetDataEngine
from cpn_core.models.plate_info import PlateInfo
from cpn_core.models.violation_detail import ViolationDetail
from cpn_core.types.api import ApiEnum
from cpn_core.types.vehicle_type import (
    VehicleTypeEnum,
    get_vehicle_enum,
)

RESPONSE_DATETIME_FORMAT: LiteralString = "%H:%M, %d/%m/%Y"

API_CAPTCHA: LiteralString = "https://www.csgt.vn/lib/captcha/captcha.class.php"
API_QUERY_1: LiteralString = "https://www.csgt.vn/?mod=contact&task=tracuu_post&ajax"
API_QUERY_2: LiteralString = "https://www.csgt.vn/tra-cuu-phuong-tien-vi-pham.html?&LoaiXe={vehicle_type}&BienKiemSoat={plate}"

logger = getLogger(__name__)


# https://github.com/PyGithub/PyGithub/issues/2300
SSL_CONTEXT: Final[SSLContext] = ssl_create_context()
SSL_CONTEXT.set_ciphers("DEFAULT@SECLEVEL=1")


class _GetDataCsgtCoreEngine:
    api: ApiEnum = ApiEnum.csgt_vn

    def __init__(self, plate_info: PlateInfo, *, timeout: float) -> None:
        self._plate_info: PlateInfo = plate_info
        self._vehicle_type: VehicleTypeEnum = get_vehicle_enum(self._plate_info.type)
        self._violations_details_set: set[ViolationDetail] = set()
        self._timeout: float = timeout
        self._session: ClientSession = ClientSession(timeout=ClientTimeout(timeout))

    @staticmethod
    def _bypass_captcha(captcha_img: bytes) -> str:
        with Image.open(BytesIO(captcha_img)) as image:
            return image_to_string(image).strip()

    async def _get_phpsessid_and_captcha(self) -> tuple[str, bytes]:
        logger.debug(f"Plate {self._plate_info.plate}: Getting cookies and captcha...")
        async with self._session.get(
            API_CAPTCHA,
            ssl=SSL_CONTEXT,
        ) as response:
            response.raise_for_status()
            phpsessid: str = response.cookies["PHPSESSID"].value
            captcha_img: bytes = await response.read()
            logger.debug(f"Plate {self._plate_info.plate} PHPSESSID: {phpsessid}")
            return phpsessid, captcha_img

    async def _get_html_check(self, captcha: str, phpsessid: str) -> str:
        payload: dict[str, str | int] = {
            "BienKS": self._plate_info.plate,
            "Xe": self._vehicle_type.value,
            "captcha": captcha,
            "ipClient": "9.9.9.91",
            "cUrl": self._vehicle_type.value,
        }
        headers: dict[str, str] = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        cookies: SimpleCookie = SimpleCookie(f"PHPSESSID={phpsessid}")
        async with self._session.post(
            url=API_QUERY_1,
            headers=headers,
            cookies=cookies,
            data=payload,
            ssl=SSL_CONTEXT,
        ) as response:
            return await response.text()

    async def _get_plate_data(self) -> str:
        async with self._session.post(
            url=API_QUERY_2.format(
                vehicle_type=self._vehicle_type.value,
                plate=self._plate_info.plate,
            ),
            ssl=SSL_CONTEXT,
        ) as response:
            return await response.text()

    def _parse_violation(self, violation_data: str) -> None:
        soup: BeautifulSoup = BeautifulSoup(violation_data, "html.parser")
        if not soup.css:
            return
        plate: str | None = (
            plate_tag.text.strip()
            if (
                plate_tag := soup.select_one(
                    ".form-group:nth-child(1) > div > div:nth-child(2)"
                )
            )
            else None
        )
        color: str | None = (
            color_tag.text.strip()
            if (
                color_tag := soup.select_one(
                    ".form-group:nth-child(2) > div > div:nth-child(2)"
                )
            )
            else None
        )
        type: str | None = (
            type_tag.text.strip()
            if (
                type_tag := soup.select_one(
                    ".form-group:nth-child(3) > div > div:nth-child(2)"
                )
            )
            else None
        )
        date: str | None = (
            date_tag.text.strip()
            if (
                date_tag := soup.select_one(
                    ".form-group:nth-child(4) > div > div:nth-child(2)"
                )
            )
            else None
        )
        location: str | None = (
            location_tag.text.strip()
            if (
                location_tag := soup.select_one(
                    ".form-group:nth-child(5) > div > div:nth-child(2)"
                )
            )
            else None
        )
        violation: str | None = (
            action_tag.text.strip()
            if (
                action_tag := soup.select_one(
                    ".form-group:nth-child(6) > div > div:nth-child(2)"
                )
            )
            else None
        )
        status: str | None = (
            status_tag.text.strip()
            if (
                status_tag := soup.select_one(
                    ".form-group:nth-child(7) > div > div:nth-child(2)"
                )
            )
            else None
        )
        enforcement_unit: str | None = (
            enforcement_unit_tag.text.strip()
            if (
                enforcement_unit_tag := soup.select_one(
                    ".form-group:nth-child(8) > div > div:nth-child(2)"
                )
            )
            else None
        )
        resolution_offices: list[str] = [
            resolution_offices_tag.text.strip()
            for resolution_offices_tag in soup.select(".form-group:nth-child(n+9)")
        ]
        if (
            plate is None
            or color is None
            or date is None
            or location is None
            or violation is None
            or status is None
            or enforcement_unit is None
            or not resolution_offices
        ):
            logger.error(
                f"Plate {self._plate_info.plate}: Cannot parse a violation data"
            )
            return
        violation_detail: ViolationDetail = ViolationDetail(
            plate=plate,
            color=color,
            type=get_vehicle_enum(type),
            # Have to cast to string because lsp's warning
            date=datetime.strptime(str(date), RESPONSE_DATETIME_FORMAT),
            location=location,
            violation=violation,
            status=status == "Đã xử phạt",
            enforcement_unit=enforcement_unit,
            resolution_offices=tuple(resolution_offices),
        )
        self._violations_details_set.add(violation_detail)

    def _parse_violations(
        self, violations_data: list[str]
    ) -> tuple[ViolationDetail, ...]:
        for violation_data in violations_data:
            self._parse_violation(violation_data)
        violation_details: tuple[ViolationDetail, ...] = tuple(
            self._violations_details_set
        )
        if not violation_details:
            logger.info(f"Plate {self._plate_info.plate}: Don't find any violation")
        return violation_details

    def _parse_html(self, html: str) -> tuple[ViolationDetail, ...] | None:
        soup: BeautifulSoup = BeautifulSoup(html, "html.parser")
        violation_group_tag: Tag | NavigableString | None = soup.find(
            "div", id="bodyPrint123"
        )
        if not violation_group_tag or isinstance(violation_group_tag, NavigableString):
            logger.error('"Cannot get the div whose id is "bodyPrint123"')
            return
        violation_group: str = violation_group_tag.prettify(formatter=None)
        # HACK: This split is hard. Maybe change it to regex split later
        violations_data: list[str] = "".join(violation_group.splitlines()[1:-2]).split(
            '<hr style="margin-bottom: 25px;"/>'
        )
        return self._parse_violations(violations_data)

    async def get_data(self) -> tuple[ViolationDetail, ...] | None:
        try:
            phpsessid, captcha_img = await self._get_phpsessid_and_captcha()
            captcha: str = self._bypass_captcha(captcha_img)
            logger.debug(f"Plate {self._plate_info.plate} captcha resolved: {captcha}")
            logger.debug(
                f"Plate {self._plate_info.plate}: Sending request again to get check..."
            )
            html_data: str = await self._get_html_check(captcha, phpsessid)
            if html_data.strip() == "404":
                logger.error(f"Plate {self._plate_info.plate}: Wrong captcha")
                return
            logger.debug(
                f"Plate {self._plate_info.plate}: Sending request again to get data..."
            )
            plate_data: str = await self._get_plate_data()
            violations: tuple[ViolationDetail, ...] | None = self._parse_html(
                plate_data
            )
            return violations
        except TimeoutError as e:
            logger.error(
                f"Plate {self._plate_info.plate}: Time out ({self._timeout}s) getting data from API {self.api.value}. {e}"
            )
        except ClientError as e:
            logger.error(
                f"Plate {self._plate_info.plate}: Error occurs while getting data from API {self.api.value}. {e}"
            )
        except Exception as e:
            logger.error(
                f"Plate {self._plate_info.plate}: Error occurs while getting data (internally) {self.api.value}. {e}"
            )

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None:
        await self._session.close()


class CsgtGetDataEngine(BaseGetDataEngine):
    def __init__(self, *, timeout: float) -> None:
        self._timeout: float = timeout
        super().__init__()

    @override
    async def get_data(
        self, plate_info: PlateInfo
    ) -> tuple[ViolationDetail, ...] | None:
        async with _GetDataCsgtCoreEngine(
            plate_info, timeout=self._timeout
        ) as local_engine:
            return await local_engine.get_data()



from enum import IntEnum
from typing import Any, Literal, TypeAlias


class VehicleTypeEnum(IntEnum):
    car = 1
    motorbike = 2
    electric_motorbike = 3


VehicleStrVieType: TypeAlias = Literal["Ô tô", "Xe máy", "Xe máy điện"]

VehicleStrType: TypeAlias = Literal["car", "motorbike", "electric_motorbike"]

VehicleIntType: TypeAlias = Literal[1, 2, 3]

VehicleType: TypeAlias = VehicleIntType | VehicleStrType | VehicleStrVieType


def get_vehicle_enum(type: VehicleTypeEnum | VehicleType | Any) -> VehicleTypeEnum:
    if isinstance(type, VehicleTypeEnum):
        return type
    match type:
        case "car" | "Ô tô" | 1 | VehicleTypeEnum.car:
            return VehicleTypeEnum.car
        case "motorbike" | "Xe máy" | 2 | VehicleTypeEnum.motorbike:
            return VehicleTypeEnum.motorbike
        case (
            "electric_motorbike"
            | "Xe máy điện"
            | 3
            | VehicleTypeEnum.electric_motorbike
        ):
            return VehicleTypeEnum.electric_motorbike
        case _:
            raise ValueError("Unknown vehicle type")


def get_vehicle_str(type: VehicleTypeEnum | VehicleType | Any) -> VehicleStrType:
    match type:
        case "car" | "Ô tô" | 1 | VehicleTypeEnum.car:
            return "car"
        case "motorbike" | "Xe máy" | 2 | VehicleTypeEnum.motorbike:
            return "motorbike"
        case (
            "electric_motorbike"
            | "Xe máy điện"
            | 3
            | VehicleTypeEnum.electric_motorbike
        ):
            return "electric_motorbike"
        case _:
            raise ValueError("Unknown vehicle type")


def get_vehicle_str_vie(type: VehicleTypeEnum | VehicleType | Any) -> VehicleStrVieType:
    match type:
        case "car" | "Ô tô" | 1 | VehicleTypeEnum.car:
            return "Ô tô"
        case "motorbike" | "Xe máy" | 2 | VehicleTypeEnum.motorbike:
            return "Xe máy"
        case (
            "electric_motorbike"
            | "Xe máy điện"
            | 3
            | VehicleTypeEnum.electric_motorbike
        ):
            return "Xe máy điện"
        case _:
            raise ValueError("Unknown vehicle type")


__all__ = [
    "VehicleIntType",
    "VehicleStrType",
    "VehicleStrVieType",
    "VehicleType",
    "VehicleTypeEnum",
    "get_vehicle_enum",
    "get_vehicle_str",
]



from abc import abstractmethod
from logging import getLogger
from typing import Self, TypeVar

from cpn_core.notifications.models.base import BaseNotificationEngineConfig

logger = getLogger(__name__)


T: TypeVar = TypeVar("T", bound=BaseNotificationEngineConfig)


class BaseNotificationEngine[T]:
    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None: ...

    @abstractmethod
    async def send(
        self,
        config: T,
        messages: tuple[str, ...],
    ) -> None: ...



from datetime import datetime
from typing import Literal, LiteralString, override

from pydantic import BaseModel

from cpn_core.constants.datetime import DATETIME_FORMAT_12, DATETIME_FORMAT_24
from cpn_core.types.vehicle_type import VehicleTypeEnum, get_vehicle_str_vie
from cpn_core.utils.gen_map_search_url import gen_map_search_url


class ViolationDetail(BaseModel):
    plate: str | None
    color: str | None
    type: VehicleTypeEnum | None
    date: datetime | None
    location: str | None
    violation: str | None
    status: bool | None
    enforcement_unit: str | None
    resolution_offices: tuple[str, ...] | None

    def get_str(
        self,
        *,
        show_less_detail: bool,
        markdown: bool,
        time_format: Literal["12", "24"],
    ) -> str:
        if markdown:
            return self._get_markdown_message(show_less_detail, time_format)
        else:
            return self._get_raw_messages(show_less_detail, time_format)

    def _get_raw_messages(
        self,
        show_less_detail: bool,
        time_format: Literal["12", "24"],
    ) -> str:
        header: LiteralString = "⚠️ Thông tin vi phạm:"
        message: str = (
            (
                (f"Biển: {self.plate}" if self.plate is not None else "")
                + (f"\nMàu biển: {self.color}" if self.color is not None else "")
                + (
                    f"\nLoại xe: {get_vehicle_str_vie(self.type)}"
                    if self.type is not None
                    else ""
                )
                + "\n"
                if not show_less_detail
                else ""
            )
            + (
                f"Thời điểm: {self.date.strftime(DATETIME_FORMAT_24) if time_format == '24' else self.date.strftime(DATETIME_FORMAT_12)}"
                if self.date is not None
                else ""
            )
            + (f"\nVị trí: {self.location}" if self.location is not None else "")
            + (f"\nHành vi: {self.violation}" if self.violation is not None else "")
            + (
                f"\nTrạng thái: {'Đã xử phạt' if self.status else 'Chưa xử phạt'}"
                if self.status is not None
                else ""
            )
            + (
                f"\nĐơn vị phát hiện: {self.enforcement_unit}"
                if self.enforcement_unit is not None
                else ""
            )
            + (
                (
                    "\n"
                    + "\n".join(
                        f"- {resolution_office}"
                        for resolution_office in self.resolution_offices
                    )
                )
                if self.resolution_offices is not None
                else ""
            )
        ).strip()
        return f"{header}\n{message}".strip()

    def _get_markdown_message(
        self,
        show_less_detail: bool,
        time_format: Literal["12", "24"],
    ) -> str:
        header: LiteralString = "**⚠️ Thông tin vi phạm:**"
        message: str = (
            (
                (f"**Biển:** {self.plate}" if self.plate is not None else "")
                + (f"\n**Màu biển:** {self.color}" if self.color is not None else "")
                + (
                    f"\n**Loại xe:** {get_vehicle_str_vie(self.type)}"
                    if self.type is not None
                    else ""
                )
                + "\n"
                if not show_less_detail
                else ""
            )
            + (
                f"**Thời điểm:** {self.date.strftime(DATETIME_FORMAT_24) if time_format == '24' else self.date.strftime(DATETIME_FORMAT_12)}"
                if self.date is not None
                else ""
            )
            + (
                f"\n**Vị trí:** [{self.location}]({gen_map_search_url(self.location)})"
                if self.location is not None
                else ""
            )
            + (f"\n**Hành vi:** {self.violation}" if self.violation else "")
            + (
                f"\n**Trạng thái:** {'Đã xử phạt ✅' if self.status else 'Chưa xử phạt ❌'}"
                if self.status is not None
                else ""
            )
            + (
                f"\n**Đơn vị phát hiện:** {self.enforcement_unit}"
                if self.enforcement_unit is not None
                else ""
            )
            + (
                (
                    "\n"
                    + "\n".join(
                        f"- {resolution_office}"
                        for resolution_office in self.resolution_offices
                    )
                )
                if self.resolution_offices is not None
                else ""
            )
        ).strip()
        return f"{header}\n{message}".strip()

    @override
    def __hash__(self):
        return (
            hash(self.plate) + hash(self.color) + hash(self.date) + hash(self.location)
        )



import asyncio
from logging import getLogger
from typing import LiteralString, override

from aiohttp import ClientError, ClientSession, ClientTimeout

from cpn_core.notifications.models.telegram import (
    TelegramNotificationEngineConfig,
)

from .base import BaseNotificationEngine

API_URL: LiteralString = "https://api.telegram.org/bot{bot_token}/sendMessage"


logger = getLogger(__name__)


class TelegramNotificationEngine(
    BaseNotificationEngine[TelegramNotificationEngineConfig]
):
    def __init__(self, *, timeout: float) -> None:
        self._timeout: float = timeout
        self._session: ClientSession = ClientSession(
            timeout=ClientTimeout(timeout),
        )

    async def _send_message(
        self,
        telegram: TelegramNotificationEngineConfig,
        message: str,
    ) -> None:
        url: str = API_URL.format(bot_token=telegram.bot_token)
        payload: dict[str, str] = {
            "chat_id": telegram.chat_id,
            "text": message,
            "parse_mode": "Markdown",
        }
        try:
            async with self._session.post(
                url,
                json=payload,
            ) as response:
                response.raise_for_status()
            logger.info(f"Successfully sent to Telegram Chat ID: {telegram.chat_id}")
        except TimeoutError as e:
            logger.error(
                f"Timeout ({self._timeout}s) sending to Telegram Chat ID: {telegram.chat_id}. {e}"
            )
            raise
        except ClientError as e:
            logger.error(f"Failed to sent to Telegram Chat ID: {telegram.chat_id}. {e}")
            raise
        except Exception as e:
            logger.error(
                f"Failed to sent to Telegram Chat ID (internally): {telegram.chat_id}. {e}"
            )
            raise

    @override
    async def send(
        self,
        config: TelegramNotificationEngineConfig,
        messages: tuple[str, ...],
    ) -> None:
        await asyncio.gather(
            *(
                self._send_message(
                    telegram=config,
                    message=message,
                )
                for message in messages
            )
        )

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None:
        await self._session.close()



from abc import abstractmethod
from logging import getLogger
from typing import Self

from cpn_core.models.plate_info import PlateInfo
from cpn_core.models.violation_detail import ViolationDetail

logger = getLogger(__name__)


class BaseGetDataEngine:
    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None: ...

    @abstractmethod
    async def get_data(
        self, plate_info: PlateInfo
    ) -> tuple[ViolationDetail, ...] | None: ...



import re
from asyncio import TimeoutError
from datetime import datetime
from logging import getLogger
from re import DOTALL
from typing import LiteralString, override

from aiohttp import ClientError, ClientSession, ClientTimeout
from bs4 import BeautifulSoup, ResultSet, Tag

from cpn_core.get_data.engines.base import BaseGetDataEngine
from cpn_core.models.plate_info import PlateInfo
from cpn_core.models.violation_detail import ViolationDetail
from cpn_core.types.api import ApiEnum
from cpn_core.types.vehicle_type import (
    get_vehicle_enum,
)

logger = getLogger(__name__)

API_URL: LiteralString = "https://api.phatnguoi.vn/web/tra-cuu"
RESPONSE_DATETIME_FORMAT: LiteralString = "%H:%M, %d/%m/%Y"


class _PhatNguoiGetDataParseEngine:
    def __init__(self, plate_info: PlateInfo, html_data: str) -> None:
        self._plate_info: PlateInfo = plate_info
        self._html_data: str = html_data
        self._violations_details_set: set[ViolationDetail] = set()

    def _parse_violation(self, violation_html: Tag) -> None:
        plate: str | None = (
            plate_tag.text.strip()
            if (
                plate_tag := violation_html.select_one(
                    "tr:nth-child(1) > td:nth-child(2)"
                )
            )
            else None
        )
        color: str | None = (
            color_tag.text.strip()
            if (
                color_tag := violation_html.select_one(
                    "tr:nth-child(2) > td:nth-child(2)"
                )
            )
            else None
        )
        type: str | None = (
            type_tag.text.strip()
            if (
                type_tag := violation_html.select_one(
                    "tr:nth-child(3) > td:nth-child(2)"
                )
            )
            else None
        )
        date: str | None = (
            date_tag.text.strip()
            if (
                date_tag := violation_html.select_one(
                    "tr:nth-child(4) > td:nth-child(2)"
                )
            )
            else None
        )
        location: str | None = (
            location_tag.text.strip()
            if (
                location_tag := violation_html.select_one(
                    "tr:nth-child(5) > td:nth-child(2)"
                )
            )
            else None
        )
        violation: str | None = (
            violation_tag.text.strip()
            if (
                violation_tag := violation_html.select_one(
                    "tr:nth-child(6) > td:nth-child(2)"
                )
            )
            else None
        )
        status: str | None = (
            status_tag.text.strip()
            if (
                status_tag := violation_html.select_one(
                    "tr:nth-child(7) > td:nth-child(2)"
                )
            )
            else None
        )
        enforcement_unit: str | None = (
            enforcement_unit_tag.text.strip()
            if (
                enforcement_unit_tag := violation_html.select_one(
                    "tr:nth-child(8) > td:nth-child(2)"
                )
            )
            else None
        )
        resolution_offices: str | None = (
            resolution_offices_tag.text.strip()
            if (
                resolution_offices_tag := violation_html.select_one(
                    "tr:nth-child(9) > td:nth-child(2)"
                )
            )
            else None
        )
        if (
            plate is None
            or color is None
            or date is None
            or location is None
            or violation is None
            or status is None
            or enforcement_unit is None
            or not resolution_offices
        ):
            logger.error(
                f"Plate {self._plate_info.plate}: Cannot parse a violation data"
            )
            return
        # # TODO: Split resolution_office as other api
        self._violations_details_set.add(
            ViolationDetail(
                plate=plate,
                color=color,
                type=get_vehicle_enum(type),
                location=location,
                # Have to cast to string because lsp's warning
                date=datetime.strptime(str(date), RESPONSE_DATETIME_FORMAT),
                violation=violation,
                status=status == "Đã xử phạt",
                enforcement_unit=enforcement_unit,
                resolution_offices=tuple(
                    re.findall(r"\d\..*?(?=(?:\d\.|$))", resolution_offices, DOTALL)
                ),
            )
        )

    def parse(self) -> tuple[ViolationDetail, ...] | None:
        soup: BeautifulSoup = BeautifulSoup(self._html_data, "html.parser")
        if not soup.css:
            logger.error("The data got doesn't have css selector ability")
            return
        violation_htmls: ResultSet[Tag] | None = soup.css.select("tbody")
        if not violation_htmls:
            logger.error("Cannot get the tbody tag")
            return
        for violation_html in violation_htmls:
            self._parse_violation(violation_html)
        violation_details: tuple[ViolationDetail, ...] = tuple(
            self._violations_details_set
        )
        if not violation_details:
            logger.info(f"Plate {self._plate_info.plate}: Don't find any violation")
        return violation_details


class PhatNguoiGetDataEngine(BaseGetDataEngine):
    api: ApiEnum = ApiEnum.phatnguoi_vn

    def __init__(self, *, timeout: float) -> None:
        self._timeout: float = timeout
        self._session: ClientSession = ClientSession(
            timeout=ClientTimeout(timeout),
        )

    async def _request(self, plate_info: PlateInfo) -> str | None:
        url: str = f"{API_URL}/{plate_info.plate}/{get_vehicle_enum(plate_info.type)}"
        try:
            async with self._session.get(url=url) as response:
                html_data: str = await response.text()
            return html_data
        except TimeoutError as e:
            logger.error(
                f"Plate {plate_info.plate}: Time out ({self._timeout}s) getting data from API {self.api.value}. {e}"
            )
        except ClientError as e:
            logger.error(
                f"Plate {plate_info.plate}: Error occurs while getting data from API {self.api.value}. {e}"
            )
        except Exception as e:
            logger.error(
                f"Plate {plate_info.plate}: Error occurs while getting data (internally) {self.api.value}. {e}"
            )

    @override
    async def get_data(
        self, plate_info: PlateInfo
    ) -> tuple[ViolationDetail, ...] | None:
        html_data: str | None = await self._request(plate_info)
        if html_data is None:
            return
        violations: tuple[ViolationDetail, ...] | None = _PhatNguoiGetDataParseEngine(
            plate_info=plate_info,
            html_data=html_data,
        ).parse()
        return violations

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None:
        await self._session.close()



from asyncio import TimeoutError
from datetime import datetime
from logging import getLogger
from typing import Literal, LiteralString, TypedDict, cast, override

from aiohttp import ClientError, ClientSession
from aiohttp.client import ClientTimeout

from cpn_core.get_data.engines.base import BaseGetDataEngine
from cpn_core.models.plate_info import PlateInfo
from cpn_core.models.violation_detail import ViolationDetail
from cpn_core.types.api import ApiEnum
from cpn_core.types.vehicle_type import get_vehicle_enum

logger = getLogger(__name__)

RESPONSE_DATETIME_FORMAT: LiteralString = "%H:%M, %d/%m/%Y"
API_URL: LiteralString = (
    "https://api.zm.io.vn/v1/csgt/tracuu?licensePlate={plate}&vehicleType={type}"
)


class _DataPlateInfoResponse(TypedDict):
    bienkiemsoat: str
    maubien: str
    loaiphuongtien: Literal["Ô tô", "Xe máy", "Xe máy điện"]
    thoigianvipham: str
    diadiemvipham: str
    trangthai: str
    donviphathienvipham: str
    noigiaiquyetvuviec: str


class _DataResponse(TypedDict):
    json: tuple[_DataPlateInfoResponse, ...] | None
    html: str
    css: str


class _Response(TypedDict):
    time_end: int
    data: _DataResponse
    error: bool


class _ZMIOGetDataParseEngine:
    def __init__(
        self,
        plate_info: PlateInfo,
        data: tuple[_DataPlateInfoResponse, ...],
    ) -> None:
        self._plate_info: PlateInfo = plate_info
        self._data: tuple[_DataPlateInfoResponse, ...] = data
        self._violations_details_set: set[ViolationDetail] = set()

    def _parse_violation(self, data: _DataPlateInfoResponse) -> None:
        plate: str = data["bienkiemsoat"]
        date: str = data["thoigianvipham"]
        type: Literal["Ô tô", "Xe máy", "Xe máy điện"] = data["loaiphuongtien"]
        color: str = data["maubien"]
        location: str = data["diadiemvipham"]
        status: str = data["trangthai"]
        enforcement_unit: str = data["donviphathienvipham"]
        # NOTE: this api just responses 1 resolution_office
        resolution_offices: tuple[str, ...] = (data["noigiaiquyetvuviec"],)
        violation_detail: ViolationDetail = ViolationDetail(
            plate=plate,
            color=color,
            type=get_vehicle_enum(type),
            date=datetime.strptime(str(date), RESPONSE_DATETIME_FORMAT),
            location=location,
            status=status == "Đã xử phạt",
            enforcement_unit=enforcement_unit,
            resolution_offices=resolution_offices,
            violation=None,
        )
        self._violations_details_set.add(violation_detail)

    def parse(self) -> tuple[ViolationDetail, ...] | None:
        for violations in self._data:
            self._parse_violation(violations)
        return tuple(self._violations_details_set)


class ZMIOGetDataEngine(BaseGetDataEngine):
    api = ApiEnum.zm_io_vn

    def __init__(self, *, timeout: float) -> None:
        self._timeout: float = timeout
        self._session: ClientSession = ClientSession(
            timeout=ClientTimeout(timeout),
        )

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None:
        await self._session.close()

    async def _request(self, plate_info: PlateInfo) -> dict | None:
        url: str = API_URL.format(
            plate=plate_info.plate, type=get_vehicle_enum(plate_info.type)
        )
        try:
            async with self._session.get(url) as response:
                json: dict = await response.json()
                return json
        except TimeoutError as e:
            logger.error(
                f"Plate {plate_info.plate}: Time out ({self._timeout}s) getting data from API {self.api.value}. {e}"
            )
        except ClientError as e:
            logger.error(
                f"Plate {plate_info.plate}: Error occurs while getting data from API {self.api.value}. {e}"
            )
        except Exception as e:
            logger.error(
                f"Plate {plate_info.plate}: Error occurs while getting data (internally) {self.api.value}. {e}"
            )

    @override
    async def get_data(
        self, plate_info: PlateInfo
    ) -> tuple[ViolationDetail, ...] | None:
        plate_detail_raw: dict | None = await self._request(plate_info)
        if not plate_detail_raw:
            return
        plate_detail_typed: _Response = cast(_Response, plate_detail_raw)
        if plate_detail_typed["data"] is None:
            logger.error(f"Plate {plate_info.plate}: Cannot get data")
            return
        if plate_detail_typed["data"]["json"] is None:
            logger.info(
                f"Plate {plate_info.plate}: Not found or don't have any violations"
            )
            return ()
        violation_details: tuple[ViolationDetail, ...] | None = _ZMIOGetDataParseEngine(
            plate_info=plate_info, data=plate_detail_typed["data"]["json"]
        ).parse()
        return violation_details



from datetime import datetime
from typing import Any, Literal, LiteralString, override

from pydantic import BaseModel, Field, computed_field

from cpn_core.constants.datetime import DATETIME_FORMAT_12, DATETIME_FORMAT_24
from cpn_core.models.plate_info import PlateInfo
from cpn_core.models.violation_detail import ViolationDetail


class PlateDetail(BaseModel):
    plate_info: PlateInfo = Field(
        description="Thông tin biển số phương tiện",
    )
    violations: tuple[ViolationDetail, ...] | None = Field(
        description="Danh sách các vi phạm của 1 biển xe",
    )
    date_time: datetime = Field(
        description="Thời gian lấy thông tin",
        default_factory=datetime.now,
    )

    @computed_field
    @property
    def total_fines(self) -> int:
        if self.violations is None:
            return -1
        return len(self.violations)

    @computed_field
    @property
    def total_peding_fines(self) -> int:
        if self.violations is None:
            return -1
        return len(
            tuple(violation for violation in self.violations if not violation.status)
        )

    def _get_fines_total_str(
        self, *, show_less_detail: bool, markdown: bool
    ) -> str | None:
        if self.violations is None:
            return None
        if markdown:
            return (
                ""
                if not show_less_detail
                else f"**Số vi phạm:** `{self.total_fines}`\n"
            ) + f"**Số vi phạm chưa xử phạt:** `{self.total_peding_fines}`"
        else:
            return (
                "" if not show_less_detail else f"Số vi phạm: {self.total_fines}\n"
            ) + (f"Số vi phạm chưa xử phạt: {self.total_peding_fines}")

    def get_str(
        self,
        *,
        show_less_detail: bool,
        markdown: bool,
        time_format: Literal["12", "24"],
    ) -> str:
        plate_info: str = self.plate_info.get_str(
            show_less_detail=show_less_detail, markdown=markdown
        )
        if self.violations is None:
            return plate_info
        fines_total: str | None = self._get_fines_total_str(
            show_less_detail=show_less_detail, markdown=markdown
        )
        violation_header: LiteralString = (
            "**Vi phạm thứ #{order}:**\n" if markdown else "Vi phạm thứ #{order}:\n"
        )
        violations: tuple[str, ...] = tuple(
            violation.get_str(
                show_less_detail=show_less_detail,
                markdown=markdown,
                time_format=time_format,
            )
            for violation in self.violations
        )
        return (
            plate_info
            + (f"\n{fines_total}" if fines_total else "")
            + "\n\n"
            + "\n\n".join(
                violation_header.format(order=order) + violation
                for order, violation in enumerate(violations, start=1)
            )
        )

    def get_messages(
        self,
        *,
        show_less_detail: bool,
        markdown: bool,
        time_format: Literal["12", "24"],
    ) -> tuple[str, ...]:
        if not self.violations:
            return ()
        plate_info: str = self.plate_info.get_str(
            show_less_detail=show_less_detail, markdown=markdown
        )
        violations: tuple[str, ...] = tuple(
            violation.get_str(
                show_less_detail=show_less_detail,
                markdown=markdown,
                time_format=time_format,
            )
            for violation in self.violations
        )
        date_time: str = self.date_time.strftime(
            DATETIME_FORMAT_24 if time_format == "24" else DATETIME_FORMAT_12
        )
        return tuple(
            f"{plate_info}\n\n{violation}\n\nGửi lúc: {date_time}"
            for violation in violations
        )

    @override
    def __hash__(self):
        return hash(self.plate_info) + hash(self.violations)

    @override
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, PlateDetail):
            return self.plate_info == other.plate_info and (
                all(x == y for x, y in zip(self.violations, other.violations))
                if self.violations and other.violations
                else (not self.violations and not other.violations)
            )
        return False


__all__ = ["PlateDetail"]



import json
from asyncio import TimeoutError
from datetime import datetime
from logging import getLogger
from typing import Final, Literal, LiteralString, TypeAlias, TypedDict, cast, override

from aiohttp import ClientError, ClientSession, ClientTimeout

from cpn_core.get_data.engines.base import BaseGetDataEngine
from cpn_core.models.plate_info import PlateInfo
from cpn_core.models.violation_detail import ViolationDetail
from cpn_core.types.api import ApiEnum
from cpn_core.types.vehicle_type import (
    VehicleStrVieType,
    VehicleTypeEnum,
    get_vehicle_enum,
)

API_URL: LiteralString = "https://api.checkphatnguoi.vn/phatnguoi"
RESPONSE_DATETIME_FORMAT: LiteralString = "%H:%M, %d/%m/%Y"

logger = getLogger(__name__)

_DataResponse = TypedDict(
    "_DataResponse",
    {
        "Biển kiểm soát": str,
        "Màu biển": str,
        "Loại phương tiện": Literal["Ô tô", "Xe máy", "Xe máy điện"],
        "Thời gian vi phạm": str,
        "Địa điểm vi phạm": str,
        "Hành vi vi phạm": str,
        "Trạng thái": Literal["Đã xử phạt", "Chưa xử phạt"],
        "Đơn vị phát hiện vi phạm": str,
        "Nơi giải quyết vụ việc": tuple[str, ...],
    },
)

_DataPlateInfoResponse = TypedDict(
    "_DataPlateInfoResponse",
    {
        "total": int,
        "chuaxuphat": Literal[0, 1],
        "daxuphat": Literal[0, 1],
        "latest": str,
    },
)

_FoundResponse = TypedDict(
    "_FoundResponse",
    {
        "status": Literal[1],
        "msg": str,
        "data": tuple[_DataResponse, ...],
    },
)


_NotFoundResponse = TypedDict(
    "_NotFoundResponse",
    {
        "status": Literal[2],
        "data": None,
    },
)

_Response: TypeAlias = _FoundResponse | _NotFoundResponse


class _CheckPhatNguoiGetDataParseEngine:
    def __init__(self, plate_info: PlateInfo, plate_detail_dict: _Response) -> None:
        self._plate_info: PlateInfo = plate_info
        self._vehicle_type: VehicleTypeEnum = get_vehicle_enum(plate_info.type)
        self._plate_detail_typed: _Response = plate_detail_dict
        self._violations_details_set: set[ViolationDetail] = set()

    def _parse_violation(self, data: _DataResponse) -> None:
        type: VehicleStrVieType = data["Loại phương tiện"]
        # NOTE: this is for filtering the vehicle that doesn't match the plate info type. Because checkphatnguoi.vn return all of the type of the plate
        parsed_type: VehicleTypeEnum = get_vehicle_enum(type)
        if parsed_type != self._vehicle_type:
            logger.info("The violation doesn't match the input vehicle type")
            return
        plate: str = data["Biển kiểm soát"]
        date: str = data["Thời gian vi phạm"]
        color: str = data["Màu biển"]
        location: str = data["Địa điểm vi phạm"]
        violation: str = data["Hành vi vi phạm"]
        status: str = data["Trạng thái"]
        enforcement_unit: str = data["Đơn vị phát hiện vi phạm"]
        resolution_offices: tuple[str, ...] = data["Nơi giải quyết vụ việc"]
        violation_detail: ViolationDetail = ViolationDetail(
            plate=plate,
            color=color,
            type=parsed_type,
            # Have to cast to string because lsp's warning
            date=datetime.strptime(str(date), RESPONSE_DATETIME_FORMAT),
            location=location,
            violation=violation,
            status=status == "Đã xử phạt",
            enforcement_unit=enforcement_unit,
            resolution_offices=resolution_offices,
        )
        self._violations_details_set.add(violation_detail)

    def parse(self) -> tuple[ViolationDetail, ...] | None:
        if self._plate_detail_typed["status"] == 2:
            logger.info("Don't have any violations")
            return ()
        if self._plate_detail_typed["status"] != 1:
            logger.error("Unknown Error with status = 1 from API")
            return
        for violation in self._plate_detail_typed["data"]:
            self._parse_violation(violation)
        return tuple(self._violations_details_set)


class CheckPhatNguoiGetDataEngine(BaseGetDataEngine):
    api: ApiEnum = ApiEnum.checkphatnguoi_vn
    headers: Final[dict[str, str]] = {"Content-Type": "application/json"}

    def __init__(self, *, timeout: float) -> None:
        self._timeout: float = timeout
        self._session: ClientSession = ClientSession(
            timeout=ClientTimeout(timeout),
        )

    async def _request(self, plate_info: PlateInfo) -> dict | None:
        payload: Final[dict[str, str]] = {"bienso": plate_info.plate}
        try:
            async with self._session.post(
                API_URL,
                headers=self.headers,
                json=payload,
            ) as response:
                response.raise_for_status()
                response_data: bytes = await response.read()
                return json.loads(response_data)
        except TimeoutError as e:
            logger.error(
                f"Plate {plate_info.plate}: Time out ({self._timeout}s) getting data from API {self.api.value}. {e}"
            )
        except ClientError as e:
            logger.error(
                f"Plate {plate_info.plate}: Error occurs while getting data from API {self.api.value}. {e}"
            )
        except Exception as e:
            logger.error(
                f"Plate {plate_info.plate}: Error occurs while getting data (internally) {self.api.value}. {e}"
            )

    @override
    async def get_data(
        self, plate_info: PlateInfo
    ) -> tuple[ViolationDetail, ...] | None:
        plate_detail_raw: dict | None = await self._request(plate_info)
        if not plate_detail_raw:
            return
        plate_detail_typed: _Response = cast(_Response, plate_detail_raw)
        violations: tuple[ViolationDetail, ...] | None = (
            _CheckPhatNguoiGetDataParseEngine(
                plate_info=plate_info, plate_detail_dict=plate_detail_typed
            ).parse()
        )
        return violations

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None:
        await self._session.close()



from typing import Any, override

from pydantic import BaseModel, ConfigDict, Field

from cpn_core.types.api import ApiEnum
from cpn_core.types.vehicle_type import (
    VehicleType,
    get_vehicle_enum,
    get_vehicle_str_vie,
)


class PlateInfo(BaseModel):
    model_config = ConfigDict(
        title="Thông tin biển số",
        frozen=True,
    )

    plate: str = Field(
        description="Biển số",
        title="Biển số",
        examples=["60A64685", "98-A-56604", "30-F88251", "59XB-00000"],
    )
    type: VehicleType = Field(
        description="Loại phương tiện để gửi request cũng như lọc loại phương tiện đối với các API không lọc loại phương tiện sẵn",
        title="Loại phương tiện",
    )
    enabled: bool = Field(
        description="Kích hoạt",
        default=True,
    )
    apis: tuple[ApiEnum, ...] | None = Field(
        description='Sử dụng API từ trang web nào. Config giống "api" ở ngoài .Để trống sẽ sử dụng API define ở scope ngoài.',
        title="API",
        default=None,
        min_length=1,
    )
    owner: str | None = Field(
        description="Ghi chú chủ sở hữu (phù hợp khi dùng nhắc ai đó với lựa chọn notifications)",
        title="Ghi chú chủ sở hữu",
        examples=["@kevinnitro", "dad"],
        default=None,
    )

    def get_str(self, *, show_less_detail: bool, markdown: bool) -> str:
        if markdown:
            return self._get_markdown_message(show_less_detail)
        else:
            return self._get_raw_messages(show_less_detail)

    def _get_raw_messages(self, show_less_detail: bool) -> str:
        message: str
        if show_less_detail:
            message = (
                (f"Biển số: {self.plate}")
                + (f"\nChủ sở hữu: {self.owner}" if self.owner else "")
            ).strip()
        else:
            message = (
                (f"Biển số: {self.plate}")
                + (f"\nChủ sở hữu: {self.owner}" if self.owner else "")
                + (f"\nLoại phương tiện: {get_vehicle_str_vie(self.type)}")
            ).strip()
        return "Thông tin phương tiện:\n" + message

    def _get_markdown_message(self, show_less_detail: bool) -> str:
        message: str
        if show_less_detail:
            message = (
                (f"*Biển số:* {self.plate}")
                + (f"\n*Chủ sở hữu:* {self.owner}" if self.owner else "")
            ).strip()
        else:
            message = (
                (f"*Biển số:* {self.plate}")
                + (f"\n*Chủ sở hữu:* {self.owner}" if self.owner else "")
                + (f"\n*Loại phương tiện:* {get_vehicle_str_vie(self.type)}")
            ).strip()
        return "*🚗 **Thông tin phương tiện**:*\n" + message

    @override
    def __hash__(self) -> int:
        return (
            hash(self.plate)
            + hash(self.type)
            + hash(self.enabled)
            + hash(self.apis)
            + hash(self.owner)
        )

    @override
    def __eq__(self, other: Any):
        if isinstance(other, PlateInfo):
            return (
                self.plate == other.plate
                and get_vehicle_enum(self.type) == get_vehicle_enum(other.type)
                and self.enabled == other.enabled
                and self.owner == other.owner
                and (
                    all(
                        x == y
                        for x, y in zip(
                            (self.apis,)
                            if isinstance(self.apis, ApiEnum)
                            else self.apis,
                            (other.apis,)
                            if isinstance(other.apis, ApiEnum)
                            else other.apis,
                        )
                    )
                    if self.apis and other.apis
                    else (not self.apis and not other.apis)
                )
            )
        return False


__all__ = ["PlateInfo"]
